import click

from . import dev
from .backend import (
    follow_logs,
    get_all_containers_by_name,
)

DEFAULT_LOG_LINES = 5


@dev.command()
@click.argument('containers', nargs=-1)
@click.option(
    'all_containers', '--all',
    is_flag=True,
    help='Show the output of all containers, not just app containers',
)
@click.option(
    '-n', '--number-of-lines',
    type=int,
    default=DEFAULT_LOG_LINES,
    help='Number of lines of history to output',
)
@click.option(
    '--with-history',
    is_flag=True,
    help='Show all of the output history rather than `-n` lines',
)
@click.pass_obj
def logs(
    kubetools_config,
    containers, all_containers,
    number_of_lines, with_history,
):
    '''
    Follow logs for the dev environment.
    '''

    click.echo(click.style(
        '--> Following dev environment logs...', 'blue',
    ))

    # Show all the top level containers
    if not containers and not all_containers:
        containers = list(get_all_containers_by_name(
            kubetools_config,
            ('deployments',),
        ).keys())

    tail = number_of_lines
    if with_history:
        # "all" is a special value for `docker-compose logs tail=X`
        tail = 'all'
        if number_of_lines != DEFAULT_LOG_LINES:
            click.echo(click.style(
                '-n={0} overridden by --with-history'.format(number_of_lines),
                'yellow',
            ))

    follow_logs(kubetools_config, containers, tail=tail)



from kubetools.deploy.util import delete_objects, log_actions
from kubetools.kubernetes.api import (
    delete_job,
    delete_namespace,
    delete_pod,
    delete_replica_set,
    get_object_name,
    is_kubetools_object,
    list_complete_jobs,
    list_namespaces,
    list_pods,
    list_replica_sets,
)


# Cleanup
# Handles removal of orphaned replicasets/pods and optionally any complete jobs,
# working on the namespace level only (no apps).
# If the cleanup removes all remaining objects, the namespace will be deleted too.

def get_cleanup_objects(build, cleanup_jobs):
    replica_sets = list_replica_sets(build.env, build.namespace)
    replica_set_names = set(get_object_name(replica_set) for replica_set in replica_sets)
    replica_sets_to_delete = []
    replica_set_names_to_delete = set()
    replica_set_names_already_deleted = set()

    for replica_set in replica_sets:
        if not is_kubetools_object(replica_set):
            continue

        if not replica_set.metadata.owner_references:
            replica_set_names_to_delete.add(get_object_name(replica_set))
            replica_sets_to_delete.append(replica_set)

        if replica_set.metadata.deletion_timestamp:
            replica_set_names_already_deleted.add(get_object_name(replica_set))

    jobs_to_delete = []
    jobs_set_names_to_delete = set()
    if cleanup_jobs:
        jobs_to_delete = list_complete_jobs(build.env, build.namespace)
        for job in jobs_to_delete:
            jobs_set_names_to_delete.add(get_object_name(job))

    pods = list_pods(build.env, build.namespace)
    pod_names = set(get_object_name(pod) for pod in pods)
    pods_to_delete = []
    pod_names_to_delete = set()
    pod_names_already_deleted = set()

    for pod in pods:
        if not pod.metadata.owner_references:
            pods_to_delete.append(pod)
            pod_names_to_delete.add(get_object_name(pod))

        elif len(pod.metadata.owner_references) == 1:
            owner = pod.metadata.owner_references[0]
            if owner.name in replica_set_names_to_delete or owner.name in jobs_set_names_to_delete:
                pods_to_delete.append(pod)
                pod_names_to_delete.add(get_object_name(pod))

        if pod.metadata.deletion_timestamp:
            pod_names_already_deleted.add(get_object_name(pod))

    namespaces = list_namespaces(build.env)
    for namespace in namespaces:
        if namespace.metadata.name == build.namespace:
            current_namespace = namespace
            break
    else:
        current_namespace = None

    namespace_to_delete = []
    # Namespace must exist. And "default" namespace can't be deleted
    if current_namespace and current_namespace.metadata.name != "default":
        remaining_pods = pod_names - pod_names_to_delete - pod_names_already_deleted
        remaining_replicasets = (
                replica_set_names - replica_set_names_to_delete - replica_set_names_already_deleted
        )

        if len(remaining_pods) == 0 and len(remaining_replicasets) == 0:
            namespace_to_delete = [current_namespace]

    return namespace_to_delete, replica_sets_to_delete, pods_to_delete, jobs_to_delete


def log_cleanup_changes(
    build, namespace, replica_sets, pods, jobs,
    message='Executing changes:',
    name_formatter=lambda name: name,
):
    with build.stage(message):
        log_actions(build, 'DELETE', 'replica_set', replica_sets, name_formatter)
        log_actions(build, 'DELETE', 'jobs', jobs, name_formatter)
        log_actions(build, 'DELETE', 'pod', pods, name_formatter)
        log_actions(build, 'DELETE', 'namespace', namespace, name_formatter)


def execute_cleanup(build, namespace, replica_sets, pods, jobs):
    with build.stage('Delete replica sets'):
        delete_objects(build, replica_sets, delete_replica_set)

    with build.stage('Delete jobs'):
        delete_objects(build, jobs, delete_job)

    with build.stage('Delete pods'):
        delete_objects(build, pods, delete_pod)

    with build.stage('Delete namespace'):
        delete_objects(build, namespace, delete_namespace)



def make_secret_volume_config(name, secret):
    volume_data = {
        'name': name,
        'csi': {
            'driver': 'secrets-store.csi.k8s.io',
            'readOnly': True,
            'volumeAttributes': {
                'secretProviderClass': secret['secretProviderClass'],
            },
        },
    }

    return volume_data



import click

from kubernetes import config

from kubetools import __version__
from kubetools.log import setup_logging
from kubetools.settings import get_settings


class SpecialHelpOrder(click.Group):
    def __init__(self, *args, **kwargs):
        self.help_priorities = {}
        super(SpecialHelpOrder, self).__init__(*args, **kwargs)

    def list_commands(self, ctx):
        '''
        Reorder the list of commands when listing the help.
        '''
        commands = super(SpecialHelpOrder, self).list_commands(ctx)
        return (
            c[1] for c in sorted(
                (self.help_priorities.get(command, 1), command)
                for command in commands
            )
        )

    def group(self, *args, **kwargs):
        '''
        Behaves the same as `click.Group.command()` except capture a priority for
        listing command names in help.
        '''

        help_priority = kwargs.pop('help_priority', 1)
        help_priorities = self.help_priorities

        def decorator(f):
            cmd = super(SpecialHelpOrder, self).group(*args, **kwargs)(f)
            help_priorities[cmd.name] = help_priority
            return cmd

        return decorator

    def command(self, *args, **kwargs):
        '''
        Behaves the same as `click.Group.command()` except capture a priority for
        listing command names in help.
        '''

        help_priority = kwargs.pop('help_priority', 1)
        help_priorities = self.help_priorities

        def decorator(f):
            cmd = super(SpecialHelpOrder, self).command(*args, **kwargs)(f)
            help_priorities[cmd.name] = help_priority
            return cmd

        return decorator


def _get_context_names():
    try:
        contexts, active_context = config.list_kube_config_contexts()
    except config.ConfigException as e:
        # The python-kubernetes library currently does not handle a missing "current context"
        # well at all, raising an exception.
        # See: https://github.com/kubernetes-client/python/issues/1193
        if 'Expected key current-context' in e.args[0]:
            raise click.ClickException((
                'No current-context set in kubeconfig! Please set this to any '
                'value using `kubectl config use-context <name>`.'
            ))
        raise

    if not contexts:
        print('Cannot find any context in kube-config file.')
        return
    return [context['name'] for context in contexts], active_context['name']


def print_contexts(ctx, param, value):
    if not value:
        return

    click.echo('--> Available Kubernetes contexts:')
    context_names, active_context_name = _get_context_names()
    for name in context_names:
        click.echo(f'    {click.style(name, bold=name == active_context_name)}')

    ctx.exit()


def ensure_context(ctx, param, value):
    context_names, active_context_name = _get_context_names()

    if value:
        if value not in context_names:
            raise click.BadParameter(f'{value}; available contexts: {context_names}')
    else:
        click.echo(f'Using active context: {click.style(active_context_name, bold=True)}')
        value = active_context_name

    return value


@click.group(cls=SpecialHelpOrder)
@click.option(
    '--context',
    callback=ensure_context,
    envvar='KUBETOOLS_CONTEXT',
    help='The name of the Kubernetes context to use.',
)
@click.option(
    '--contexts',
    is_flag=True,
    is_eager=True,
    callback=print_contexts,
    expose_value=False,
    help='List available Kubernetes contexts and exit.',
)
@click.option('--debug', is_flag=True, help='Show debug logs.')
@click.version_option(version=__version__, message='%(prog)s: v%(version)s')
@click.pass_context
def cli_bootstrap(ctx, context, debug):
    '''
    Kubetools client - deploy apps to Kubernetes.
    '''

    ctx.meta['kube_context'] = context

    setup_logging(debug)
    get_settings()



import sys

from kubetools.settings import get_settings

from . import docker_compose


DEV_BACKEND_PROVIDERS = {
    'docker_compose': docker_compose,
}


# Now, using settings splice in the .backend module to match the provider
settings = get_settings()

try:
    backend = DEV_BACKEND_PROVIDERS[settings.DEV_BACKEND]
except KeyError:
    raise KeyError('Invalid dev backend: {0}'.format(settings.DEV_BACKEND))

sys.modules['kubetools.dev.backend'] = backend
backend.init_backend()



from time import sleep

from kubernetes import client, config
from kubernetes.client.rest import ApiException

from kubetools.constants import MANAGED_BY_ANNOTATION_KEY
from kubetools.exceptions import KubeBuildError
from kubetools.settings import get_settings


def get_object_labels_dict(obj):
    return obj.metadata.labels or {}


def get_object_annotations_dict(obj):
    if hasattr(obj.metadata, "annotations") and obj.metadata.annotations:
        return obj.metadata.annotations
    elif hasattr(obj.metadata, "template"):
        return get_object_annotations_dict(obj.spec.template)
    else:
        return {}


def get_object_name(obj):
    if isinstance(obj, dict):
        return obj['metadata']['name']
    return obj.metadata.name


def is_kubetools_object(obj):
    if get_object_annotations_dict(obj).get(MANAGED_BY_ANNOTATION_KEY) == 'kubetools':
        return True


def namespace_exists(env, namespace_obj):
    k8s_core_api = _get_k8s_core_api(env)
    return _object_exists(k8s_core_api, 'read_namespace', None, namespace_obj)


def list_namespaces(env):
    k8s_core_api = _get_k8s_core_api(env)
    return k8s_core_api.list_namespace().items


def create_namespace(env, namespace_obj):
    k8s_core_api = _get_k8s_core_api(env)
    k8s_namespace = k8s_core_api.create_namespace(
        body=namespace_obj,
    )

    _wait_for_object(k8s_core_api, 'read_namespace', None, namespace_obj)
    return k8s_namespace


def update_namespace(env, namespace_obj):
    k8s_core_api = _get_k8s_core_api(env)
    k8s_namespace = k8s_core_api.patch_namespace(
        name=get_object_name(namespace_obj),
        body=namespace_obj,
    )

    return k8s_namespace


def delete_namespace(env, _namespace, namespace_obj):
    k8s_core_api = _get_k8s_core_api(env)
    k8s_core_api.delete_namespace(
        name=get_object_name(namespace_obj),
    )

    _wait_for_no_object(k8s_core_api, 'read_namespace', None, namespace_obj)


def list_pods(env, namespace):
    k8s_core_api = _get_k8s_core_api(env)
    return k8s_core_api.list_namespaced_pod(namespace=namespace).items


def delete_pod(env, namespace, pod):
    k8s_core_api = _get_k8s_core_api(env)
    k8s_core_api.delete_namespaced_pod(
        name=get_object_name(pod),
        namespace=namespace,
    )

    _wait_for_no_object(k8s_core_api, 'read_namespaced_pod', namespace, pod)


def list_replica_sets(env, namespace):
    k8s_apps_api = _get_k8s_apps_api(env)
    return k8s_apps_api.list_namespaced_replica_set(namespace=namespace).items


def delete_replica_set(env, namespace, replica_set):
    k8s_apps_api = _get_k8s_apps_api(env)
    k8s_apps_api.delete_namespaced_replica_set(
        name=get_object_name(replica_set),
        namespace=namespace,
    )

    _wait_for_no_object(k8s_apps_api, 'read_namespaced_replica_set', namespace, replica_set)


def list_services(env, namespace):
    k8s_core_api = _get_k8s_core_api(env)
    return k8s_core_api.list_namespaced_service(namespace=namespace).items


def delete_service(env, namespace, service):
    k8s_core_api = _get_k8s_core_api(env)
    k8s_core_api.delete_namespaced_service(
        name=get_object_name(service),
        namespace=namespace,
    )

    _wait_for_no_object(k8s_core_api, 'read_namespaced_service', namespace, service)


def service_exists(env, namespace, service):
    k8s_core_api = _get_k8s_core_api(env)
    return _object_exists(k8s_core_api, 'read_namespaced_service', namespace, service)


def create_service(env, namespace, service):
    k8s_core_api = _get_k8s_core_api(env)
    k8s_service = k8s_core_api.create_namespaced_service(
        body=service,
        namespace=namespace,
    )

    _wait_for_object(k8s_core_api, 'read_namespaced_service', namespace, service)
    return k8s_service


def update_service(env, namespace, service):
    k8s_core_api = _get_k8s_core_api(env)
    k8s_service = k8s_core_api.patch_namespaced_service(
        name=get_object_name(service),
        body=service,
        namespace=namespace,
    )

    return k8s_service


def list_deployments(env, namespace):
    k8s_apps_api = _get_k8s_apps_api(env)
    return k8s_apps_api.list_namespaced_deployment(namespace=namespace).items


def delete_deployment(env, namespace, deployment):
    k8s_apps_api = _get_k8s_apps_api(env)
    k8s_apps_api.delete_namespaced_deployment(
        name=get_object_name(deployment),
        namespace=namespace,
    )

    _wait_for_no_object(k8s_apps_api, 'read_namespaced_deployment', namespace, deployment)


def deployment_exists(env, namespace, deployment):
    k8s_apps_api = _get_k8s_apps_api(env)
    return _object_exists(k8s_apps_api, 'read_namespaced_deployment', namespace, deployment)


def create_deployment(env, namespace, deployment):
    k8s_apps_api = _get_k8s_apps_api(env)
    k8s_deployment = k8s_apps_api.create_namespaced_deployment(
        body=deployment,
        namespace=namespace,
    )

    wait_for_deployment(env, namespace, k8s_deployment)
    return k8s_deployment


def update_deployment(env, namespace, deployment):
    k8s_apps_api = _get_k8s_apps_api(env)
    k8s_deployment = k8s_apps_api.patch_namespaced_deployment(
        name=get_object_name(deployment),
        body=deployment,
        namespace=namespace,
    )

    wait_for_deployment(env, namespace, k8s_deployment)
    return k8s_deployment


def wait_for_deployment(env, namespace, deployment):
    k8s_apps_api = _get_k8s_apps_api(env)

    def check_deployment():
        d = k8s_apps_api.read_namespaced_deployment(
            name=get_object_name(deployment),
            namespace=namespace,
        )

        if d.status.ready_replicas == d.status.replicas:
            return True

    _wait_for(check_deployment, get_object_name(deployment))


def list_cronjobs(env, namespace):
    _batch_api_version, k8s_batch_api = _get_compatible_cronjob_api(env)
    return k8s_batch_api.list_namespaced_cron_job(namespace=namespace).items


def delete_cronjob(env, namespace, cronjob):
    _batch_api_version, k8s_batch_api = _get_compatible_cronjob_api(env)
    k8s_batch_api.delete_namespaced_cron_job(
        name=get_object_name(cronjob),
        namespace=namespace,
    )

    _wait_for_no_object(k8s_batch_api, 'read_namespaced_cron_job', namespace, cronjob)


def cronjob_exists(env, namespace, cronjob):
    job_api_version = _get_cronjob_api_version(cronjob)
    batch_api_version, k8s_batch_api = _get_compatible_cronjob_api(env, job_api_version)
    cronjob['apiVersion'] = batch_api_version
    return _object_exists(k8s_batch_api, 'read_namespaced_cron_job', namespace, cronjob)


def create_cronjob(env, namespace, cronjob):
    job_api_version = _get_cronjob_api_version(cronjob)
    batch_api_version, k8s_batch_api = _get_compatible_cronjob_api(env, job_api_version)
    cronjob['apiVersion'] = batch_api_version
    k8s_cronjob = k8s_batch_api.create_namespaced_cron_job(
        body=cronjob,
        namespace=namespace,
    )

    return k8s_cronjob


def update_cronjob(env, namespace, cronjob):
    job_api_version = _get_cronjob_api_version(cronjob)
    batch_api_version, k8s_batch_api = _get_compatible_cronjob_api(env, job_api_version)
    cronjob['apiVersion'] = batch_api_version
    k8s_cronjob = k8s_batch_api.patch_namespaced_cron_job(
        name=get_object_name(cronjob),
        body=cronjob,
        namespace=namespace,
    )

    return k8s_cronjob


def list_jobs(env, namespace):
    k8s_batch_api = _get_k8s_jobs_batch_api(env)
    return k8s_batch_api.list_namespaced_job(namespace=namespace).items


def list_running_jobs(env, namespace):
    jobs = list_jobs(env, namespace)
    return [job for job in jobs if _is_job_running(job)]


def list_complete_jobs(env, namespace):
    jobs = list_jobs(env, namespace)
    return [job for job in jobs if not _is_job_running(job)]


def _is_job_running(job):
    conditions = job.status.conditions
    if conditions is None:
        return True
    complete = any(condition.type == 'Complete' for condition in job.status.conditions)
    return not complete


valid_propagation_policies = ["Orphan", "Background", "Foreground"]


def delete_job(env, namespace, job, propagation_policy=None):
    if propagation_policy and propagation_policy not in valid_propagation_policies:
        raise KubeBuildError(f"Propagation policy must be one of {valid_propagation_policies}")
    args = {}
    if propagation_policy:
        args['propagation_policy'] = propagation_policy
    k8s_batch_api = _get_k8s_jobs_batch_api(env)
    k8s_batch_api.delete_namespaced_job(
        name=get_object_name(job),
        namespace=namespace,
        **args,
    )

    _wait_for_no_object(k8s_batch_api, 'read_namespaced_job', namespace, job)


def create_job(env, namespace, job):
    k8s_batch_api = _get_k8s_jobs_batch_api(env)
    k8s_job = k8s_batch_api.create_namespaced_job(
        body=job,
        namespace=namespace,
    )

    wait_for_job(env, namespace, k8s_job)
    return k8s_job


def wait_for_job(env, namespace, job):
    k8s_batch_api = _get_k8s_jobs_batch_api(env)

    def check_job():
        j = k8s_batch_api.read_namespaced_job(
            name=get_object_name(job),
            namespace=namespace,
        )

        if j.status.succeeded == j.spec.completions:
            return True

    _wait_for(check_job, get_object_name(job))


def _wait_for_object(*args):
    return _wait_for(lambda: _object_exists(*args) is True)


def _wait_for_no_object(*args):
    return _wait_for(lambda: _object_exists(*args) is False)


def _wait_for(function, name='object'):
    settings = get_settings()

    sleeps = 0
    while True:
        if function():
            return

        sleep(settings.WAIT_SLEEP_TIME)
        sleeps += 1

        if sleeps > settings.WAIT_MAX_SLEEPS:
            raise KubeBuildError(f'Timeout waiting for {name} to be ready')


def _object_exists(api, method, namespace, obj):
    try:
        if namespace:
            getattr(api, method)(
                namespace=namespace,
                name=get_object_name(obj),
            )
        else:
            getattr(api, method)(
                name=get_object_name(obj),
            )
    except ApiException as e:
        if e.status == 404:
            return False
        raise
    return True


def _get_cronjob_api_version(cronjob_obj):
    default_version = get_settings().CRONJOBS_BATCH_API_VERSION
    return cronjob_obj.get('apiVersion', default_version)


def _get_compatible_cronjob_api(env, default_version=None):
    possible_versions = [
        'batch/v1',
        'batch/v1beta1',
    ]
    if default_version:
        possible_versions.insert(0, default_version)

    for version in possible_versions:
        batch_api = _get_k8s_cronjobs_batch_api(env, version)
        api_resources = batch_api.get_api_resources()
        if any([r.name == 'cronjobs' for r in api_resources.resources]):
            return version, batch_api

    raise ApiException(
        'The target cluster does not support any Cronjob version handled by kubetools',
    )


def _get_k8s_jobs_batch_api(env):
    api_client = _get_api_client(env)
    return client.BatchV1Api(api_client=api_client)


def _get_k8s_cronjobs_batch_api(env, batch_api_version):
    api_client = _get_api_client(env)
    if batch_api_version == 'batch/v1':
        return client.BatchV1Api(api_client=api_client)
    elif batch_api_version == 'batch/v1beta1':
        return client.BatchV1beta1Api(api_client=api_client)
    else:
        raise ApiException(
            'Kubetools does not yet handle Cronjob with "batch-api-version: {0}".'
            .format(batch_api_version),
        )


def _get_k8s_core_api(env):
    api_client = _get_api_client(env)
    return client.CoreV1Api(api_client=api_client)


def _get_k8s_apps_api(env):
    api_client = _get_api_client(env)
    return client.AppsV1Api(api_client=api_client)


def _get_api_client(env):
    return config.new_client_from_config(context=env)



import os

from subprocess import CalledProcessError, check_output, STDOUT

from kubetools.constants import NAME_LABEL_KEY, PROJECT_NAME_LABEL_KEY
from kubetools.exceptions import KubeBuildError
from kubetools.kubernetes.api import (
    get_object_labels_dict,
    get_object_name,
    is_kubetools_object,
)
from kubetools.log import logger


def run_shell_command(*command, **kwargs):
    '''
    Run a shell command and return it's output. Capture fails and pass to internal
    exception.
    '''

    cwd = kwargs.pop('cwd', None)
    env = kwargs.pop('env', {})

    new_env = os.environ.copy()
    new_env.update(env)

    logger.debug(f'Running shell command in {cwd}: {command}, env: {env}')

    try:
        return check_output(command, stderr=STDOUT, cwd=cwd, env=new_env)

    except CalledProcessError as e:
        raise KubeBuildError('Command failed: {0}\n\n{1}'.format(
            ' '.join(command),
            e.output.decode('utf-8', 'ignore'),
        ))


def log_actions(build, action, object_type, names, name_formatter):
    for name in names:
        if not isinstance(name, str):
            name = get_object_name(name)
        build.log_info(f'{action} {object_type} {name_formatter(name)}')


def delete_objects(build, objects, delete_function):
    for obj in objects:
        build.log_info(f'Delete: {get_object_name(obj)}')
        delete_function(build.env, build.namespace, obj)


def get_app_objects(
    build, app_or_project_names, list_objects_function,
    force=False,
):
    objects = list_objects_function(build.env, build.namespace)

    def filter_object(obj):
        if not is_kubetools_object(obj):
            if force:
                warning = f'Will touch {get_object_name(obj)} that is not managed by kubetools!'
            else:
                warning = f'Refusing to touch {get_object_name(obj)} as not managed by kubetools!'

            build.log_warning(warning)
            return force is True
        return True

    objects = list(filter(filter_object, objects))

    if app_or_project_names:
        matched_object_names = set()

        def filter_object_names(obj):
            labels = get_object_labels_dict(obj)
            app_name = labels.get(NAME_LABEL_KEY)
            if app_name in app_or_project_names:
                matched_object_names.add(app_name)
                return True

            project_name = labels.get(PROJECT_NAME_LABEL_KEY)
            if project_name in app_or_project_names:
                matched_object_names.add(project_name)
                return True

            return False

        objects = list(filter(filter_object_names, objects))

    return objects



import logging
import sys

import click

STDOUT_LOG_LEVELS = (logging.DEBUG, logging.INFO)
STDERR_LOG_LEVELS = (logging.WARNING, logging.ERROR, logging.CRITICAL)


logger = logging.getLogger('kubetools')


class LogFilter(logging.Filter):
    def __init__(self, *levels):
        self.levels = levels

    def filter(self, record):
        return record.levelno in self.levels


class LogFormatter(logging.Formatter):
    level_to_format = {
        logging.DEBUG: lambda s: click.style(s, 'green'),
        logging.WARNING: lambda s: click.style(s, 'yellow'),
        logging.ERROR: lambda s: click.style(s, 'red'),
        logging.CRITICAL: lambda s: click.style(s, 'red', bold=True),
    }

    def format(self, record):
        message = record.msg
        if record.args:
            message = record.msg % record.args

        # We only handle strings here
        if isinstance(message, str):
            if record.levelno in self.level_to_format:
                message = self.level_to_format[record.levelno](message)

            return message

        # If not a string, pass to standard Formatter
        else:
            return super(LogFormatter, self).format(record)


def setup_logging(debug=False):
    log_level = 'WARNING'

    if debug:
        log_level = 'DEBUG'

    # Set the log level
    logger.setLevel(getattr(logging, log_level))

    # Setup a new handler for stdout & stderr
    stdout_handler = logging.StreamHandler(sys.stdout)
    stderr_handler = logging.StreamHandler(sys.stderr)

    # Setup filters to push different levels to different streams
    stdout_filter = LogFilter(*STDOUT_LOG_LEVELS)
    stdout_handler.addFilter(stdout_filter)

    stderr_filter = LogFilter(*STDERR_LOG_LEVELS)
    stderr_handler.addFilter(stderr_filter)

    # Setup a formatter
    formatter = LogFormatter()
    stdout_handler.setFormatter(formatter)
    stderr_handler.setFormatter(formatter)

    # Add the handlers
    logger.addHandler(stdout_handler)
    logger.addHandler(stderr_handler)

    logger.debug('Log level: {0}'.format(log_level))



import json

from collections import defaultdict

import click
import yaml

from kubetools.config import load_kubetools_config
from kubetools.kubernetes.api import get_object_name
from kubetools.kubernetes.config import generate_kubernetes_configs_for_project

from . import cli_bootstrap


yaml.Dumper.ignore_aliases = lambda *args: True

FORMATTERS = {
    'json': lambda d: json.dumps(d, indent=4),
    'yaml': lambda d: yaml.dump(d),
}


@cli_bootstrap.command(help_priority=4)
@click.option(
    '--replicas',
    type=int,
    default=1,
    help='Default number of replicas for each app.',
)
@click.option(
    '-f', '--file',
    nargs=1,
    help='Specify a non-default Kubetools yml file to generate config from.',
    type=click.Path(exists=True),
)
@click.option(
    '--format', 'output_format',
    type=click.Choice(('json', 'yaml')),
    default='json',
    help='Specify the output format',
)
@click.argument(
    'app_dir',
    type=click.Path(exists=True, file_okay=False),
)
@click.option(
    '--default-registry',
    help='Default registry for apps that do not specify.',
)
@click.pass_context
def config(ctx, replicas, file, app_dir, output_format, default_registry):
    '''
    Generate and write out Kubernetes configs for a project.
    '''

    env = ctx.meta['kube_context']
    kubetools_config = load_kubetools_config(app_dir, env=env, custom_config_file=file)
    context_to_image = defaultdict(lambda: 'IMAGE')
    services, deployments, jobs, cronjobs = generate_kubernetes_configs_for_project(
        kubetools_config,
        replicas=replicas,
        context_name_to_image=context_to_image,
        default_registry=default_registry,
    )

    echo_resources(services, 'Service', output_format)
    echo_resources(deployments, 'Deployment', output_format)
    echo_resources(jobs, 'Job', output_format)
    echo_resources(cronjobs, 'Cronjob', output_format)


def echo_resources(resources, resource_kind, output_format):
    formatter = FORMATTERS[output_format]
    for resource in resources:
        name = get_object_name(resource)
        click.echo(f'{resource_kind}: {click.style(name, bold=True)}')
        click.echo(formatter(resource))
        click.echo()



from configparser import ConfigParser
from functools import lru_cache
from os import access, environ, listdir, path, X_OK

import click

from .log import logger


class KubetoolsSettings(object):
    DEFAULT_KUBE_ENV = 'staging'  # default environment when speaking to the server

    DEV_DEFAULT_ENV = 'dev'  # default environment name in dev
    DEV_HOST = 'localhost'  # dev host to link people to (should map to 127.0.0.1)
    DEV_CONFIG_DIRNAME = '.kubetools'  # project directroy to generate compose config
    DEV_BACKEND = 'docker_compose'  # backend to use for development

    CRONJOBS_BATCH_API_VERSION = 'batch/v1'  # if k8s version < 1.21+ should be 'batch/v1beta1'

    REGISTRY_CHECK_SCRIPT = None
    ''' Optional external script to check if an image exists in the docker registry.

    For cases when the registry cannot be checked by just using the docker registry V2 API.

    The script will be passed 3 arguments:
    * the registry IP/hostname and port (in the form `<ip_or_hostname>:<port>`)
    * the image name, which comes from the app name
    * the image tag (version), which comes from the build context and commit hash
    Concretely they could be used in `docker pull <registry>/<image_name>:<image_tag>`

    The script must return one of the following codes:
    * 0: the image was found in the registry
    * 1: the image was not found in the registry
    * 2: the image should be checked with the HTTP Docker V2 API
    * anything else: the script ran into an error and `kubetools` must abort
    Note that return code 2 is useful for example if the script is only responsible for checking
    some combination of registries or images, but not all.
    '''

    WAIT_SLEEP_TIME = 3
    WAIT_MAX_TIME = int(environ.get('KUBETOOLS_WAIT_MAX_TIME', 300))
    WAIT_MAX_SLEEPS = WAIT_MAX_TIME / WAIT_SLEEP_TIME

    def __init__(self, filename=None):
        self.filename = filename
        self.scripts = []


def get_settings_directory():
    return click.get_app_dir('kubetools', force_posix=True)


@lru_cache(maxsize=1)
def get_settings():
    settings_directory = get_settings_directory()
    settings_file = path.join(settings_directory, 'kubetools.conf')

    settings = KubetoolsSettings(filename=settings_file)

    if path.exists(settings_file):
        logger.info('Loading settings file: {0}'.format(settings_file))
        parser = ConfigParser()
        parser.read(settings_file)

        for option in parser.options('kubetools'):
            setattr(
                settings,
                option.upper().replace('-', '_'),
                parser.get('kubetools', option),
            )

    else:
        logger.info('No settings file: {0}'.format(settings_file))

    scripts_directory = path.join(settings_directory, 'scripts')
    if path.exists(scripts_directory):
        for filename in listdir(scripts_directory):
            if access(path.join(scripts_directory, filename), X_OK):
                settings.scripts.append(filename)

    return settings



from unittest import TestCase

from kubetools.dev.backends.docker_compose import build_run_compose_command


class TestBuildComposeCommand(TestCase):
    def setUp(self):
        self.container = 'container'

    def test_simple_composite_command(self):
        command = ('python', 'scripts/blah.py', '--argument')
        result = build_run_compose_command(self.container, command, envvars=None)
        expected_result = ['run', '--entrypoint', 'python scripts/blah.py --argument', 'container']
        self.assertEqual(expected_result, result)

    def test_simple_single_command(self):
        command = ('python scripts/blah.py --argument',)
        result = build_run_compose_command(self.container, command, envvars=None)
        expected_result = ['run', '--entrypoint', 'python scripts/blah.py --argument', 'container']
        self.assertEqual(expected_result, result)

    def test_composite_command_is_shell_escaped(self):
        command = ('python', 'scripts/blah.py', '--argument', 'key: value')
        result = build_run_compose_command(self.container, command, envvars=None)
        expected_result = [
            'run',
            '--entrypoint',
            "python scripts/blah.py --argument 'key: value'",
            'container',
        ]
        self.assertEqual(expected_result, result)

    def test_environment_variables_are_added(self):
        command = ('python', 'scripts/blah.py', '--argument')
        envvars = ('var1=val1', 'var2=val2')
        result = build_run_compose_command(self.container, command, envvars)
        expected_result = [
            'run',
            '-evar1=val1',
            '-evar2=val2',
            '--entrypoint',
            'python scripts/blah.py --argument',
            'container',
        ]
        self.assertEqual(expected_result, result)



#!/usr/bin/env python

import sys

import click

from .exceptions import KubeDevCommandError, KubeDevError, KubeError


def run_cli(func):
    try:
        func()

    except KubeDevCommandError as e:
        message, stdout = e.args

        click.echo('--> {0} {1}'.format(
            click.style('Kubetools dev exception:', 'red', bold=True),
            message,
        ))
        click.echo(stdout)
        sys.exit(1)

    except KubeDevError as e:
        click.echo('--> {0} {1}'.format(
            click.style('Kubetools dev exception:', 'red', bold=True),
            e,
        ))
        sys.exit(1)

    except KubeError as e:
        click.echo('--> {0} {1}'.format(
            click.style('Kubetools {0} exception:'.format(e.type), 'red', bold=True),
            e,
        ))
        sys.exit(1)

    except KeyboardInterrupt:
        click.echo()
        click.echo('Exiting on user request...')

    except Exception as e:
        click.echo('--> Unexpected exception: {0}'.format(
            click.style(
                '{0}{1}'.format(e.__class__.__name__, e.args),
                'red',
                bold=True,
            ),
        ))

        raise



from kubetools.constants import (
    MANAGED_BY_ANNOTATION_KEY,
    NAME_LABEL_KEY,
    PROJECT_NAME_LABEL_KEY,
    ROLE_LABEL_KEY,
)
from kubetools.exceptions import KubeConfigError

from .cronjob import make_cronjob_config
from .deployment import make_deployment_config
from .job import make_job_config
from .namespace import make_namespace_config
from .service import make_service_config
from .util import copy_and_update


def make_deployment_name(project_name, deployment_name):
    project_name = project_name.replace('_', '-')
    deployment_name = deployment_name.replace('_', '-')

    if deployment_name.startswith(project_name):
        return deployment_name

    # We expect to come here - but some old kubetools.yml's will prefix
    # the project name before every deployment.
    return '-'.join((project_name, deployment_name))


def make_context_name(app_name, container_name):
    return '{0}-{1}'.format(app_name, container_name)


def _ensure_image(
    container, context_name_to_image,
    deployment_name=None,
    container_name=None,
    default_registry=None,
):
    if 'image' in container:
        return

    if 'containerContext' in container:
        context_name = container.pop('containerContext')
        container.pop('build', None)

    elif 'build' in container and deployment_name and container_name:
        # Because no shared context was provided we use the deployment + container name
        context_name = make_context_name(deployment_name, container_name)
        container.pop('build')
    else:
        raise KubeConfigError('No image for container: {0}'.format(container))

    container['image'] = context_name_to_image[context_name]

    registry_image = container['image'].split('/')
    if len(registry_image) == 1 and default_registry is not None:
        # If registry is not specified but a default was provided
        container['image'] = f'{default_registry}/{container["image"]}'


def _get_containers_data(
    containers,
    context_name_to_image,
    deployment_name,
    default_registry=None,
):
    # Setup top level app service mapping all ports from all top level containers
    all_container_ports = []
    all_containers = {}

    for container_name, data in containers.items():
        all_container_ports.extend(data.pop('ports', []))
        _ensure_image(
            data, context_name_to_image,
            deployment_name, container_name,
            default_registry=default_registry,
        )
        all_containers[container_name] = data

    return all_containers, all_container_ports


def _get_replicas(deployment, default=1):
    replicas = default

    if 'replicaMultiplier' in deployment:
        replicas = replicas * deployment['replicaMultiplier']

    if 'maxReplicas' in deployment:
        replicas = min(replicas, deployment['maxReplicas'])

    if 'minReplicas' in deployment:
        replicas = max(replicas, deployment['minReplicas'])

    return replicas


def generate_namespace_config(name, base_labels=None, base_annotations=None):
    base_annotations = copy_and_update(base_annotations, {
        MANAGED_BY_ANNOTATION_KEY: 'kubetools',
    })

    base_labels = copy_and_update(base_labels, {
        PROJECT_NAME_LABEL_KEY: name,
    })

    namespace = make_namespace_config(
        name,
        labels=base_labels,
        annotations=base_annotations,
    )

    return namespace


def generate_kubernetes_configs_for_project(
    config,  # a kubetools config object
    replicas=1,  # number of replicas for each deployment
    envvars=None,  # global environment variables to inject to all containers

    # Upgrades/jobs
    include_upgrade_jobs=True,  # whether to generate jobs for config.upgrades
    job_specs=None,  # additional job configs to execute

    # Labels/annotations
    base_labels=None,
    base_annotations=None,
    per_deployment_annotations=None,  # per-deployment extra annotations

    # Map of context name -> docker image - images are expected to be built by
    # something else (normally the kubetools server or ktd client).
    context_name_to_image=None,
    default_registry=None,
):
    '''
    Builds service & deployment definitions based on the app config provided. We have
    app configs for each of our Kubernetes-ready apps, and they specify the basic spec
    minus things we want to change via Kubetools (replicas, versions).

    Returns two ordered lists (service configs, deployment configs). The builder executes
    them in-order, starting with the services. This means we can start dependencies before
    the main apps, so they lookup the correct settings.
    '''

    project_name = config['name']

    base_labels = copy_and_update(base_labels, {
        PROJECT_NAME_LABEL_KEY: project_name,
    })

    base_annotations = copy_and_update(base_annotations, {
        MANAGED_BY_ANNOTATION_KEY: 'kubetools',
    })

    envvars = copy_and_update(envvars, {
        'KUBE': 'true',
    })

    base_annotations = base_annotations or {}
    per_deployment_annotations = per_deployment_annotations or {}

    job_specs = job_specs or []

    services = []
    deployments = []

    for name, dependency in config.get('dependencies', {}).items():
        dependency_name = make_deployment_name(project_name, name)
        dependency_labels = copy_and_update(base_labels, {
            ROLE_LABEL_KEY: 'dependency',
            NAME_LABEL_KEY: dependency_name,
        })

        node_selector_labels = dependency.get('nodeSelector', None)
        service_account_name = dependency.get('serviceAccountName', None)
        secrets = dependency.get('secrets', None)

        containers, container_ports = _get_containers_data(
            dependency['containers'],
            context_name_to_image=context_name_to_image,
            deployment_name=name,
            default_registry=default_registry,
        )
        app_annotations = copy_and_update(base_annotations)

        if container_ports:
            services.append(make_service_config(
                dependency_name,
                container_ports,
                labels=dependency_labels,
                annotations=app_annotations,
            ))

        # For now, all dependencies use one replica
        deployments.append(make_deployment_config(
            dependency_name,
            containers,
            replicas=1,
            labels=dependency_labels,
            annotations=app_annotations,
            envvars=envvars,
            update_strategy=dependency.get('updateStrategy'),
            node_selector_labels=node_selector_labels,
            service_account_name=service_account_name,
            secrets=secrets,
        ))

    for name, deployment in config.get('deployments', {}).items():
        deployment_name = make_deployment_name(project_name, name)
        deployment_labels = copy_and_update(base_labels, {
            ROLE_LABEL_KEY: 'app',
            NAME_LABEL_KEY: deployment_name,
        })

        node_selector_labels = deployment.get('nodeSelector', None)
        service_account_name = deployment.get('serviceAccountName', None)
        secrets = deployment.get('secrets', None)

        containers, container_ports = _get_containers_data(
            deployment['containers'],
            context_name_to_image=context_name_to_image,
            deployment_name=name,
            default_registry=default_registry,
        )
        app_annotations = copy_and_update(
            base_annotations,
            per_deployment_annotations.get(name),
        )

        if container_ports:
            services.append(make_service_config(
                deployment_name,
                container_ports,
                labels=deployment_labels,
                annotations=app_annotations,
            ))

        # Setup top level app deployment
        deployment_replicas = _get_replicas(deployment, default=replicas)
        deployments.append(make_deployment_config(
            deployment_name,
            containers,
            replicas=deployment_replicas,
            labels=deployment_labels,
            annotations=app_annotations,
            envvars=envvars,
            update_strategy=deployment.get('updateStrategy'),
            node_selector_labels=node_selector_labels,
            service_account_name=service_account_name,
            secrets=secrets,
        ))

    # Jobs can be upgrades and/or passed in as part of the build spec
    jobs = []
    job_labels = copy_and_update(base_labels, {
        ROLE_LABEL_KEY: 'job',
    })

    # Add any upgrade jobs
    if include_upgrade_jobs:
        job_specs = config.get('upgrades', []) + job_specs

    for job_spec in job_specs:
        _ensure_image(job_spec, context_name_to_image, default_registry=default_registry)

        # Stil no image? Let's pull the first container we have available - this
        # maintains backwards compatability where one can ask for a job without
        # specifying any container (back when every app was one container).
        if 'image' not in job_spec:
            for name, data in config.get('deployments').items():
                found_image = False

                for _, container in data['containers'].items():
                    if 'image' in container:
                        job_spec['image'] = container['image']
                        found_image = True
                        break

                if found_image:
                    break

            # We did not break - no context found!
            else:
                raise KubeConfigError((
                    'Could not find a containerContext to use for job: {0}'
                ).format(job_spec))

        job_envvars = copy_and_update(
            envvars,
            job_spec.get('envars'),  # legacy support TODO: remove!
            job_spec.get('envvars'),
        )

        node_selector_labels = job_spec.get('nodeSelector', None)
        service_account_name = job_spec.get('serviceAccountName', None)
        secrets = job_spec.get('secrets', None)

        jobs.append(make_job_config(
            job_spec,
            app_name=project_name,
            labels=job_labels,
            annotations=base_annotations,
            envvars=job_envvars,
            node_selector_labels=node_selector_labels,
            service_account_name=service_account_name,
            secrets=secrets,
        ))

    cronjobs = []

    for name, cronjob in config.get('cronjobs', {}).items():
        cronjob_labels = copy_and_update(base_labels, {
            ROLE_LABEL_KEY: 'cronjob',
            NAME_LABEL_KEY: name,
        })

        node_selector_labels = cronjob.get('nodeSelector', None)
        service_account_name = cronjob.get('serviceAccountName', None)
        secrets = cronjob.get('secrets', None)

        containers, container_ports = _get_containers_data(
            cronjob['containers'],
            context_name_to_image=context_name_to_image,
            deployment_name=name,
            default_registry=default_registry,
        )

        app_annotations = copy_and_update(base_annotations)
        schedule = cronjob['schedule']
        concurrency_policy = cronjob['concurrency_policy']
        batch_api_version = cronjob.get('batch-api-version')  # May depend on target cluster

        cronjobs.append(make_cronjob_config(
            config,
            name,
            schedule,
            batch_api_version,
            concurrency_policy,
            containers,
            labels=cronjob_labels,
            annotations=app_annotations,
            envvars=envvars,
            node_selector_labels=node_selector_labels,
            service_account_name=service_account_name,
            secrets=secrets,
        ))

    return services, deployments, jobs, cronjobs



import shlex
from uuid import uuid4

from .container import make_container_config
from .util import copy_and_update
from .volume import make_secret_volume_config


def make_job_config(
    config,
    app_name=None,
    labels=None,
    annotations=None,
    envvars=None,
    job_name=None,
    container_name="upgrade",
    node_selector_labels=None,
    service_account_name=None,
    secrets=None,
):
    '''
    Builds a Kubernetes job configuration dict.
    '''

    # We want a copy of these because we'll modify them below
    labels = labels or {}
    envvars = envvars or {}
    annotations = annotations or {}

    # Generate name
    job_id = str(uuid4())
    if job_name is None:
        job_name = job_id

    # Attach the ID to labels
    labels = copy_and_update(labels, {
        'job-id': job_id,
    })

    # Figure out the command
    command = config['command']

    if isinstance(command, str):
        command = shlex.split(command)

    # Get/create description
    description = config.get('description', 'Run: {0}'.format(command))

    # Attach description to annotations
    annotations = copy_and_update(annotations, {
        'description': description,
    })

    # Update global envvars with job specific ones
    envvars = copy_and_update(
        envvars,
        config.get('envars'),  # legacy support TODO: remove!
        config.get('envvars'),
        {'KUBE_JOB_ID': job_id},
    )

    # Make our container
    container = make_container_config(
        job_id,
        {
            'name': container_name,
            'command': command,
            'image': config['image'],
            'chdir': config.get('chdir', '/'),
            'resources': config.get('resources', {}),
        },
        envvars=envvars,
        labels=labels,
        annotations=annotations,
        secrets=secrets,
    )

    # Completions default to 1, same as Kubernetes
    completions = config.get('completions', 1)
    # Parallelism defaults to completions, also as Kubernetes
    parallelism = config.get('parallelism', completions)

    template_spec = {
        'restartPolicy': 'Never',
        'containers': [container],
    }

    if node_selector_labels is not None:
        template_spec['nodeSelector'] = node_selector_labels

    if service_account_name is not None:
        template_spec['serviceAccountName'] = service_account_name

    if secrets is not None:
        kubernetes_volumes = []
        for secret_name, secret in secrets.items():
            kubernetes_volumes.append(make_secret_volume_config(
                secret_name, secret,
            ))
        template_spec['volumes'] = kubernetes_volumes

    job_config = {
        # Normal Kubernetes job config
        'apiVersion': 'batch/v1',
        'kind': 'Job',
        'metadata': {
            'name': job_name,
            'labels': labels,
            'annotations': annotations,
        },
        'spec': {
            'completions': completions,
            'parallelism': parallelism,
            'selector': labels,
            'template': {
                'metadata': {
                    'labels': labels,
                },
                'spec': template_spec,
            },
        },
    }

    if 'ttl_seconds_after_finished' in config:
        job_config['spec']['ttlSecondsAfterFinished'] = config.get('ttl_seconds_after_finished')

    return job_config



import shlex

from .container import make_container_config
from .util import copy_and_update
from .volume import make_secret_volume_config


def make_cronjob_config(
    config,
    cronjob_name,
    schedule,
    batch_api_version,
    concurrency_policy,
    containers,
    labels=None,
    annotations=None,
    envvars=None,
    node_selector_labels=None,
    service_account_name=None,
    secrets=None,
):
    '''
    Builds a Kubernetes cronjob configuration dict.
    '''

    labels = labels or {}
    annotations = annotations or {}

    # Build our container list
    kubernetes_containers = []
    for container_name, container in containers.items():
        # Figure out the command
        command = container['command']
        if isinstance(command, str):
            command = shlex.split(command)

        # Get/create description
        description = config.get('description', 'Run: {0}'.format(command))

        # Attach description to annotations
        annotations = copy_and_update(annotations, {
            'description': description,
        })

        kubernetes_containers.append(make_container_config(
            container_name, container,
            envvars=envvars,
            labels=labels,
            annotations=annotations,
            secrets=secrets,
        ))

    template_spec = {
        'restartPolicy': 'OnFailure',
        'containers': kubernetes_containers,
    }

    if node_selector_labels is not None:
        template_spec['nodeSelector'] = node_selector_labels

    if service_account_name is not None:
        template_spec['serviceAccountName'] = service_account_name

    if secrets is not None:
        kubernetes_volumes = []
        for secret_name, secret in secrets.items():
            kubernetes_volumes.append(make_secret_volume_config(
                secret_name, secret,
            ))
        template_spec['volumes'] = kubernetes_volumes

    # The actual cronjob spec
    cronjob = {
        'kind': 'CronJob',
        'metadata': {
            'name': cronjob_name,
            'labels': labels,
            'annotations': annotations,
        },
        'spec': {
            'schedule': schedule,
            'startingDeadlineSeconds': 10,
            'concurrencyPolicy': concurrency_policy,
            'jobTemplate': {
                'spec': {
                    'template': {
                        'metadata': {
                            'name': cronjob_name,
                            'labels': labels,
                            'annotations': annotations,
                        },
                        'spec': template_spec,
                    },
                },
            },
        },
    }
    if batch_api_version:
        # Only set here if user has specified it in the config
        cronjob['apiVersion'] = batch_api_version

    return cronjob



import os

import click

from kubetools.cli import cli_bootstrap
from kubetools.cli.git_utils import get_git_info
from kubetools.config import load_kubetools_config
from kubetools.deploy.build import Build
from kubetools.deploy.image import ensure_docker_images


@cli_bootstrap.command(help_priority=5)
@click.option(
    '--default-registry',
    help='Default registry for apps that do not specify.',
)
@click.option(
    'additional_tags', '-t', '--tag',
    multiple=True,
    help='Extra tags to apply to built image',
)
@click.option(
    'build_args', '-b', '--build-arg',
    multiple=True,
    help='Arguments to pass to docker build (Dockerfile ARG) as ARG=VALUE',
)
@click.pass_context
def push(ctx, default_registry, additional_tags, build_args):
    '''
    Push app images to docker repo
    '''
    build = Build(
        env=ctx.meta['kube_context'],
        namespace=None,
    )
    app_dir = os.getcwd()
    commit_hash, _ = get_git_info(app_dir)
    kubetools_config = load_kubetools_config(
        app_dir,
        env=build.env,
        namespace=build.namespace,
        app_name=app_dir,
        custom_config_file=False,
    )
    ensure_docker_images(
        kubetools_config, build, app_dir,
        commit_hash=commit_hash,
        default_registry=default_registry,
        additional_tags=additional_tags,
        build_args=build_args,
    )



import json
import os

import click

from kubetools.cli import cli_bootstrap
from kubetools.deploy.build import Build
from kubetools.deploy.commands.cleanup import (
    execute_cleanup,
    get_cleanup_objects,
    log_cleanup_changes,
)
from kubetools.deploy.commands.deploy import (
    execute_deploy,
    get_deploy_objects,
    log_deploy_changes,
)
from kubetools.deploy.commands.remove import (
    execute_remove,
    get_remove_objects,
    log_remove_changes,
)
from kubetools.deploy.commands.restart import (
    execute_restart,
    get_restart_objects,
    log_restart_changes,
)
from kubetools.kubernetes.api import get_object_name


def _dry_deploy_object_loop(object_type, objects):
    name_to_object = {
        get_object_name(obj): obj
        for obj in objects
    }

    while True:
        object_name = click.prompt(
            f'Print {object_type}?',
            type=click.Choice(name_to_object),
            default='exit',
        )

        if object_name == 'exit':
            break

        click.echo(json.dumps(name_to_object[object_name], indent=4))


def _dry_deploy_loop(build, namespace, services, deployments, jobs, cronjobs):
    for object_type, objects in (
        ('service', services),
        ('deployment', deployments),
        ('job', jobs),
        ('cronjob', cronjobs),
    ):
        if objects:
            _dry_deploy_object_loop(object_type, objects)


def _validate_key_value_argument(ctx, param, value):
    key_values = {}

    for key_value_str in value:
        try:
            key, value = key_value_str.split('=', 1)
        except ValueError:
            raise click.BadParameter(f'"{key_value_str}" does not match "key=value".')
        else:
            key_values[key] = value

    return key_values


@cli_bootstrap.command(help_priority=0)
@click.option(
    '--dry',
    is_flag=True,
    default=False,
    help='Instead of writing objects to Kubernetes, just print them and exit.',
)
@click.option(
    '--replicas',
    type=int,
    help='Default number of replicas for each app.',
)
@click.option(
    '--default-registry',
    help='Default registry for apps that do not specify.',
)
@click.option(
    '-y', '--yes',
    is_flag=True,
    default=False,
    help='Flag to auto-yes remove confirmation step.',
)
@click.option(
    'build_args', '-b', '--build-arg',
    multiple=True,
    help='Arguments to pass to docker build (Dockerfile ARG) as ARG=VALUE',
)
@click.option(
    'envvars', '-e', '--envvar',
    multiple=True,
    callback=_validate_key_value_argument,
    help='Extra environment variables to apply to Kubernetes objects, format: key=value.',
)
@click.option(
    'annotations', '-a', '--annotation',
    multiple=True,
    callback=_validate_key_value_argument,
    help='Extra annotations to apply to Kubernetes objects, format: key=value.',
)
@click.option(
    '-f', '--file',
    nargs=1,
    help='Specify a non-default Kubetools yml file to deploy from.',
    type=click.Path(exists=True),
)
@click.option(
    '--ignore-git-changes',
    is_flag=True,
    default=False,
    help='Flag to ignore un-committed changes in git.',
)
@click.option(
    'delete_completed_jobs', '--delete-jobs/--no-delete-jobs',
    is_flag=True,
    default=True,
    help='Delete jobs after they complete.',
)
@click.argument('namespace')
@click.argument(
    'app_dirs',
    nargs=-1,
    type=click.Path(exists=True, file_okay=False),
)
@click.pass_context
def deploy(
    ctx,
    dry,
    replicas,
    default_registry,
    yes,
    build_args,
    envvars,
    annotations,
    file,
    ignore_git_changes,
    delete_completed_jobs,
    namespace,
    app_dirs,
):
    '''
    Deploy an app, or apps, to Kubernetes.
    '''

    if not app_dirs:
        app_dirs = (os.getcwd(),)

    build = Build(
        env=ctx.meta['kube_context'],
        namespace=namespace,
    )

    if file:
        custom_config_file = click.format_filename(file)
    else:
        custom_config_file = None

    namespace, services, deployments, jobs, cronjobs = get_deploy_objects(
        build, app_dirs,
        replicas=replicas,
        default_registry=default_registry,
        build_args=build_args,
        extra_envvars=envvars,
        extra_annotations=annotations,
        ignore_git_changes=ignore_git_changes,
        custom_config_file=custom_config_file,
    )

    if not any((namespace, services, deployments, jobs, cronjobs)):
        click.echo('Nothing to do!')
        return

    if dry:
        return _dry_deploy_loop(build, namespace, services, deployments, jobs, cronjobs)

    log_deploy_changes(
        build, namespace, services, deployments, jobs, cronjobs,
        message='Executing changes:' if yes else 'Proposed changes:',
        name_formatter=lambda name: click.style(name, bold=True),
    )

    if not yes:
        click.confirm(click.style((
            'Are you sure you wish to CREATE and UPDATE the above resources? '
            'This cannot be undone.'
        )), abort=True)
        click.echo()

    execute_deploy(
        build,
        namespace,
        services,
        deployments,
        jobs,
        cronjobs,
        delete_completed_jobs=delete_completed_jobs,
    )


@cli_bootstrap.command(help_priority=1)
@click.option(
    '-y', '--yes',
    is_flag=True,
    default=False,
    help='Flag to auto-yes remove confirmation step.',
)
@click.option(
    '-f', '--force',
    is_flag=True,
    default=False,
    help='Force kubetools to remove objects it does not own.',
)
@click.option(
    '--cleanup', 'do_cleanup',
    is_flag=True,
    default=False,
    help='Run a cleanup immediately after removal.',
)
@click.argument('namespace')
@click.argument('app_or_project_names', nargs=-1)
@click.pass_context
def remove(ctx, yes, force, do_cleanup, namespace, app_or_project_names):
    '''
    Removes one or more apps from a given namespace.
    '''

    build = Build(
        env=ctx.meta['kube_context'],
        namespace=namespace,
    )

    services_to_delete, deployments_to_delete, jobs_to_delete, cronjobs_to_delete = (
        get_remove_objects(build, app_or_project_names, force=force)
    )

    if not any((services_to_delete, deployments_to_delete, jobs_to_delete, cronjobs_to_delete)):
        click.echo('Nothing to do 👍!')
        return

    log_remove_changes(
        build, services_to_delete, deployments_to_delete, jobs_to_delete, cronjobs_to_delete,
        message='Executing changes:' if yes else 'Proposed changes:',
        name_formatter=lambda name: click.style(name, bold=True),
    )

    if not yes:
        click.confirm(click.style(
            'Are you sure you wish to DELETE the above resources? This cannot be undone.',
        ), abort=True)
        click.echo()

    execute_remove(
        build,
        services_to_delete,
        deployments_to_delete,
        jobs_to_delete,
        cronjobs_to_delete,
    )

    if do_cleanup:
        ctx.invoke(cleanup, yes=yes, namespace=namespace)


@cli_bootstrap.command(help_priority=2)
@click.option(
    '--cleanup-jobs',
    'cleanup_jobs',
    is_flag=True,
    default=False,
    help='Additionally cleanup completed jobs and associated pods',
)
@click.option(
    '-y', '--yes',
    is_flag=True,
    default=False,
    help='Flag to auto-yes remove confirmation step.',
)
@click.argument('namespace')
@click.pass_context
def cleanup(ctx, cleanup_jobs, yes, namespace):
    '''
    Cleans up a namespace by removing orphaned objects.
    Will delete the namespace if it's empty after cleanup.
    '''

    build = Build(
        env=ctx.meta['kube_context'],
        namespace=namespace,
    )

    namespace_to_delete, replica_sets_to_delete, pods_to_delete, jobs_to_delete =\
        get_cleanup_objects(build, cleanup_jobs)

    if not any((namespace_to_delete, replica_sets_to_delete, pods_to_delete, jobs_to_delete)):
        click.echo('Nothing to do 👍!')
        return

    log_cleanup_changes(
        build, namespace_to_delete, replica_sets_to_delete, pods_to_delete, jobs_to_delete,
        message='Executing changes:' if yes else 'Proposed changes:',
        name_formatter=lambda name: click.style(name, bold=True),
    )

    if not yes:
        click.confirm(click.style(
            'Are you sure you wish to DELETE the above resources? This cannot be undone.',
        ), abort=True)
        click.echo()

    execute_cleanup(
        build,
        namespace_to_delete,
        replica_sets_to_delete,
        pods_to_delete,
        jobs_to_delete,
    )


@cli_bootstrap.command()
@click.option(
    '-y', '--yes',
    is_flag=True,
    default=False,
    help='Flag to auto-yes remove confirmation step.',
)
@click.option(
    '-f', '--force',
    is_flag=True,
    default=False,
    help='Force kubetools to remove objects it does not own.',
)
@click.argument('namespace')
@click.argument('app_or_project_names', nargs=-1)
@click.pass_context
def restart(ctx, yes, force, namespace, app_or_project_names):
    '''
    Restarts one or more apps in a given namespace.
    '''

    build = Build(
        env=ctx.meta['kube_context'],
        namespace=namespace,
    )

    deployments_and_pods_to_delete = get_restart_objects(
        build, app_or_project_names,
        force=force,
    )

    if not deployments_and_pods_to_delete:
        click.echo('Nothing to do 👍!')
        return

    log_restart_changes(
        build, deployments_and_pods_to_delete,
        message='Executing changes:' if yes else 'Proposed changes:',
        name_formatter=lambda name: click.style(name, bold=True),
    )

    if not yes:
        click.confirm(click.style(
            'Are you sure you wish to DELETE the above resources? This cannot be undone.',
        ), abort=True)
        click.echo()

    execute_restart(
        build,
        deployments_and_pods_to_delete,
    )



from os import listdir, path
from unittest import mock, TestCase

import yaml

from kubetools.config import load_kubetools_config
from kubetools.deploy.image import get_container_contexts_from_config, get_docker_tag_for_commit
from kubetools.kubernetes.api import get_object_name
from kubetools.kubernetes.config import generate_kubernetes_configs_for_project


class TestKubernetesConfigGeneration(TestCase):
    def test_basic_app_configs(self):
        _test_configs('basic_app')

    def test_dependencies_configs(self):
        _test_configs('dependencies')

    def test_dev_overrides_configs(self):
        _test_configs('dev_overrides', dev=True)

    def test_k8s_container_passthrough_configs(self):
        _test_configs('k8s_container_passthrough')

    def test_k8s_cronjobs_beta_api_version_configs(self):
        _test_configs('k8s_cronjobs_beta_api_version')

    def test_multiple_deployments_configs(self):
        _test_configs('multiple_deployments')

    def test_docker_registry_configs(self):
        _test_configs('docker_registry', default_registry='default-registry')

    def test_k8s_with_mounted_secrets_configs(self):
        _test_configs('k8s_with_mounted_secrets')


def _test_configs(folder_name, default_registry=None, **kwargs):
    app_dir = path.join('tests', 'configs', folder_name)

    kubetools_config = load_kubetools_config(app_dir, **kwargs)

    # TODO: refactor deploy.image._ensure_docker_images to extract the logic to a function and
    #  de-duplicate it from here
    build_contexts = get_container_contexts_from_config(kubetools_config)
    context_name_to_registry = {
        context_name: build_context.get('registry', default_registry)
        for context_name, build_context in build_contexts.items()
    }
    context_images = {
        # Build the context name -> image dict
        context_name: get_docker_tag_for_commit(
            context_name_to_registry[context_name],
            kubetools_config['name'],
            context_name,
            'thisisacommithash',
        )
        for context_name in build_contexts.keys()
    }

    with mock.patch('kubetools.kubernetes.config.job.uuid4', lambda: 'UUID'):
        services, deployments, jobs, cronjobs = generate_kubernetes_configs_for_project(
            kubetools_config,
            default_registry=default_registry,
            context_name_to_image=context_images,
        )

    k8s_files = listdir(app_dir)

    if services or 'k8s_services.yml' in k8s_files:
        _assert_yaml_objects(services, path.join(app_dir, 'k8s_services.yml'))
    if deployments or 'k8s_deployments.yml' in k8s_files:
        _assert_yaml_objects(deployments, path.join(app_dir, 'k8s_deployments.yml'))
    if jobs or 'k8s_jobs.yml' in k8s_files:
        _assert_yaml_objects(jobs, path.join(app_dir, 'k8s_jobs.yml'))
    if cronjobs and 'k8s_cronjobs.yml' in k8s_files:
        _assert_yaml_objects(cronjobs, path.join(app_dir, 'k8s_cronjobs.yml'))
    if cronjobs and 'k8s_cronjobs_beta.yml' in k8s_files:
        _assert_yaml_objects(cronjobs, path.join(app_dir, 'k8s_cronjobs_beta.yml'))


def _assert_yaml_objects(objects, yaml_filename):
    with open(yaml_filename, 'r') as f:
        desired_objects = list(yaml.safe_load_all(f))

    objects.sort(key=get_object_name)
    desired_objects.sort(key=get_object_name)

    assert objects == desired_objects



from .util import get_hash

# Keys to turn into environment variables
LABEL_ENVAR_KEYS = (
    'role',
    'app',
    'name',
    'commit',
    'project_name',
    'git_name',
    'manifest_name',
)
ANNOTATION_ENVAR_KEYS = ('version',)


def _make_probe_config(config):
    if 'httpGet' in config:
        # Ensure we have a HTTP port
        if 'port' not in config['httpGet']:
            config['httpGet']['port'] = 80

        # Ensure we have a HTTP path
        if 'path' not in config['httpGet']:
            config['httpGet']['path'] = '/'

    return config


def make_container_config(
    name, container,
    envvars=None, labels=None,
    annotations=None, secrets=None,
):
    '''
    Builds the common Kubernetes container config.
    '''

    envvars = envvars or {}
    labels = labels or {}
    annotations = annotations or {}

    image = container['image']

    container_data = {
        'name': name,

        # Always pull the image from the registry
        'imagePullPolicy': 'Always',
        'image': image,

        # Environment flag we use to determine if app is in Kube
        'env': [],
    }

    # Copy these keys as-is
    for key in ('livenessProbe', 'readinessProbe'):
        if key in container:
            container_data[key] = _make_probe_config(container.pop(key))

    # Probes is a shortcut for both ready and live probes
    if 'probes' in container:
        probes = container.pop('probes')
        container_data['livenessProbe'] = _make_probe_config(probes)
        container_data['readinessProbe'] = _make_probe_config(probes)

    # Attach any of these labels as envvars
    for key in LABEL_ENVAR_KEYS:
        if key in labels:
            env_key = 'KUBETOOLS_{0}'.format(key.upper())
            container_data['env'].append({
                'name': env_key,
                'value': labels[key],
            })

    # Attach any of these annotations as envvars
    for key in ANNOTATION_ENVAR_KEYS:
        if key in annotations:
            env_key = 'KUBETOOLS_{0}'.format(key.upper())
            container_data['env'].append({
                'name': env_key,
                'value': annotations[key],
            })

    # Attach environment from the config
    if 'environment' in container:
        for item in container.pop('environment'):
            k, v = item.split('=')

            container_data['env'].append({
                'name': k,
                'value': v,
            })

    # Attach extra envvars
    if envvars:
        container_data['env'].extend([
            {
                'name': key,
                'value': str(value),
            }
            for key, value in envvars.items()
        ])

    # Apply any container-config level ENVars
    if container.get('env'):
        container_data['env'].extend([
            {
                'name': key,
                'value': str(value),
            }
            for key, value in container['env'].items()
        ])

    if 'volumes' in container:
        container_data.setdefault('volumeMounts', [])
        for volume in container.pop('volumes'):
            container_data['volumeMounts'].append({
                'mountPath': volume.split(':')[1],
                'name': get_hash(volume),
            })

    if secrets is not None:
        container_data.setdefault('volumeMounts', [])
        for secret_name, secret in secrets.items():
            container_data['volumeMounts'].append({
                'name': secret_name,
                'mountPath': secret.get('mountPath'),
                'readonly': True,
            })

    # Finally, attach all remaining data
    container_data.update(container)

    return container_data



import os

from collections import deque
from time import sleep

import click


# Hacky way of getting terminal size (so can clear lines)
# Source: http://stackoverflow.com/questions/566746
TERMINAL_SIZE = os.popen('stty size', 'r').read().split()
IS_TTY = bool(TERMINAL_SIZE)
TERMINAL_WIDTH = int(TERMINAL_SIZE[1]) if IS_TTY else None

# Rotate the spinner every 1/UPDATE_DIVISOR seconds
UPDATE_DIVISOR = 20

# Get stdout as defined by Click
STDOUT = click.get_text_stream('stdout')


def _clear_line(return_=True):
    line = ''.join(' ' for _ in range(0, TERMINAL_WIDTH))

    if return_:
        line = '{0}\r'.format(line)

    STDOUT.write(line)
    STDOUT.flush()


def wait_with_spinner(
    func,
    check_status_divisor=UPDATE_DIVISOR,
    tick_divisor=UPDATE_DIVISOR,
):
    wait_chars = deque(('-', '/', '|', '\\'))
    wait_ticks = 0

    # Store previous status so we still print progress w/o a tty, but only when
    # it changes rather than every .05s.
    previous_status = ''

    while True:
        # Get build state every ~1s
        if wait_ticks % check_status_divisor == 0:
            status = func(previous_status)

        # None = complete, so just break the loop
        if status is None:
            break

        status = status.strip()

        # Write status spinner
        if IS_TTY:
            # Limit prefix + status text width to terminal width
            width_limit = TERMINAL_WIDTH - 50
            if len(status) > width_limit:
                status = '{0}...'.format(status[:width_limit])

            _clear_line()
            STDOUT.write('  {0} in progress{1}\r'.format(
                wait_chars[0],
                ' (status = {0})'.format(status) if status else '',
            ))
            STDOUT.flush()
            wait_chars.rotate(1)

        # Print out changes to status only when no tty
        else:
            if status != previous_status:
                click.echo('    status: {0}'.format(status))

        previous_status = status

        # Tick tock
        wait_ticks += 1
        sleep(1 / tick_divisor)

    # This hack clears the progress bar line
    if IS_TTY:
        _clear_line()



from .util import make_dns_safe_name


def make_namespace_config(
    name,
    labels=None,
    annotations=None,
):
    '''
    Builds a Kubernetes namespace configuration dict.
    '''

    labels = labels or {}
    annotations = annotations or {}

    # The actual namespace spec
    namespace = {
        'apiVersion': 'v1',
        'kind': 'Namespace',
        'metadata': {
            'name': make_dns_safe_name(name),
            'labels': labels,
            'annotations': annotations,
        },
    }

    return namespace



from collections import defaultdict

from kubetools.deploy.util import get_app_objects, log_actions
from kubetools.kubernetes.api import (
    delete_pod,
    get_object_name,
    list_deployments,
    list_pods,
    list_replica_sets,
    wait_for_deployment,
)


# Restart
# Handles restarting a deployment by deleting each pod and waiting for recovery

def get_restart_objects(build, app_names=None, force=False):
    deployments = get_app_objects(
        build,
        app_names,
        list_deployments,
        force=force,
    )
    name_to_deployment = {
        get_object_name(deployment): deployment
        for deployment in deployments
    }

    replica_sets = list_replica_sets(build.env, build.namespace)
    replica_set_names_to_deployment = {}

    for replica_set in replica_sets:
        if not replica_set.metadata.owner_references:
            build.log_warning((
                'Found replicaSet with no owner (needs cleanup): '
                f'{replica_set.metadata.name}'
            ))
            continue

        if len(replica_set.metadata.owner_references) > 1:
            build.log_error((
                'Found replicaSet with more than one owner: '
                f'{replica_set.metadata.name}'
            ))
            continue

        owner_name = replica_set.metadata.owner_references[0].name
        if owner_name in name_to_deployment:
            replica_set_names_to_deployment[get_object_name(replica_set)] = (
                name_to_deployment[owner_name]
            )

    pods = list_pods(build.env, build.namespace)
    deployment_name_to_pods = defaultdict(list)

    for pod in pods:
        if pod.metadata.owner_references and len(pod.metadata.owner_references) == 1:
            owner = pod.metadata.owner_references[0]
            deployment = replica_set_names_to_deployment.get(owner.name)
            if deployment:
                deployment_name_to_pods[get_object_name(deployment)].append(pod)

    return [
        (name_to_deployment[name], pods)
        for name, pods in deployment_name_to_pods.items()
    ]


def log_restart_changes(
    build, deployments_and_pods,
    message='Executing changes:',
    name_formatter=lambda name: name,
):
    deployments = [deployment for deployment, _ in deployments_and_pods]
    with build.stage(message):
        log_actions(build, 'RESTART', 'deployment', deployments, name_formatter)


def execute_restart(build, deployments_and_pods):
    for deployment, pods in deployments_and_pods:
        with build.stage(f'Restart pods for {get_object_name(deployment)}'):
            for pod in pods:
                build.log_info(f'Delete pod: {get_object_name(pod)}')
                delete_pod(build.env, build.namespace, pod)
                wait_for_deployment(build.env, build.namespace, deployment)



from .container import make_container_config
from .util import get_hash, make_dns_safe_name
from .volume import make_secret_volume_config

DEPLOYMENT_REVISION_LIMIT = 5


def make_deployment_config(
    name, containers,
    replicas=1,
    labels=None,
    annotations=None,
    envvars=None,
    update_strategy=None,
    node_selector_labels=None,
    service_account_name=None,
    secrets=None,
):
    '''
    Builds a Kubernetes deployment configuration dict.
    '''

    labels = labels or {}
    annotations = annotations or {}

    # Build our container list
    kubernetes_containers = []
    for container_name, container in containers.items():
        kubernetes_containers.append(make_container_config(
            container_name, container,
            envvars=envvars,
            labels=labels,
            annotations=annotations,
            secrets=secrets,
        ))

    template_spec = {
        'containers': kubernetes_containers,
    }

    if node_selector_labels is not None:
        template_spec['nodeSelector'] = node_selector_labels

    if service_account_name is not None:
        template_spec['serviceAccountName'] = service_account_name

    if secrets is not None:
        kubernetes_volumes = []
        for secret_name, secret in secrets.items():
            kubernetes_volumes.append(make_secret_volume_config(
                secret_name, secret,
            ))
        template_spec['volumes'] = kubernetes_volumes

    # The actual controller Kubernetes config
    controller = {
        'apiVersion': 'apps/v1',
        'kind': 'Deployment',
        'metadata': {
            'name': make_dns_safe_name(name),
            'labels': labels,
            'annotations': annotations,
        },
        'spec': {
            'revisionHistoryLimit': DEPLOYMENT_REVISION_LIMIT,
            'selector': {
                'matchLabels': labels,
            },
            'replicas': replicas,
            'template': {
                'metadata': {
                    'labels': labels,
                },
                'spec': template_spec,
            },
        },
    }

    if update_strategy:
        controller['spec']['strategy'] = update_strategy

    container_volumes = {}
    for container in containers.values():
        if 'volumes' in container:
            for volume in container['volumes']:
                name = get_hash(volume)
                container_volumes[name] = {
                    'name': name,
                    'hostPath': {
                        'path': volume.split(':')[0],
                    },
                }

    if container_volumes:  # kube does not like an empty list
        controller['spec']['template']['spec']['volumes'] = container_volumes.values()

    for container in containers:
        if 'volumes' in container:
            for volume in container['volumes']:
                controller['spec']['template'].setdefault('volumes', []).append({
                    'name': get_hash(volume),
                    'hostPath': {
                        'path': volume.split(':')[0],
                    },
                })

    return controller



import logging
import re

from subprocess import CalledProcessError, PIPE, Popen, STDOUT
from threading import Thread

from kubetools.cli.server_util import UPDATE_DIVISOR, wait_with_spinner
from kubetools.exceptions import KubeDevCommandError
from kubetools.log import logger


def _read_command_output(command, output_lines):
    # Read the commands output indefinitely
    while True:
        stdout_line = command.stdout.readline()

        if stdout_line:
            # Remove any trailing newline
            stdout_line = stdout_line.decode().strip('\n')
            # # Strip non-alphanumeric  characters
            # stdout_line = re.sub(r'[^a-zA-Z0-9_\.\-\s]', '', stdout_line)

            # Strip any ANSI escape characters
            stdout_line = re.sub(
                r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]',
                '',
                stdout_line,
            )

            output_lines.append(stdout_line)

        # No line from the command? We're done!
        else:
            break


def _run_process_with_spinner(args):
    command = Popen(
        ' '.join(args),
        stdout=PIPE,
        stderr=STDOUT,
        close_fds=True,
        shell=True,
    )

    # Buffers
    output_lines = []

    command_reader = Thread(
        target=_read_command_output,
        args=(command, output_lines),
    )
    command_reader.start()

    def check_status(previous_status):
        # Command complete (we've read everything)? Exit here
        if not command_reader.is_alive():
            return

        if output_lines:
            return output_lines[-1]

        return previous_status

    wait_with_spinner(
        check_status,
        # This means we run the get_line check every .5 seconds
        check_status_divisor=(UPDATE_DIVISOR / 2),
    )

    # Re-join the stdout/stderr lines
    stdout = '\n'.join(output_lines)

    # Poll the command to populate it's return code
    command.poll()

    # Ensure the command is dead
    try:
        command.terminate()
        command.kill()

    # If already dead, just ignore
    except Exception:
        pass

    return command.returncode, stdout


def run_process(args, env=None, hide_output=False):
    if logger.level <= logging.DEBUG:  # always show output when debugging
        hide_output = False

    logger.debug('--> Executing: {0}'.format(' '.join(args)))

    try:
        # If we're capturing output - things are more complicated. We need to spawn
        # the subprocess in a thread and read its output into two lists, which we
        # then rejoin to return.
        if hide_output:
            code, stdout = _run_process_with_spinner(args)

        # Inline? Simply start the process and "communicate", this will print stdout
        # and stderr to the terminal and also capture them into variables.
        else:
            command = Popen(args, env=env, stderr=STDOUT, close_fds=True)
            stdout, _ = command.communicate()
            code = command.returncode

        if code > 0:
            raise KubeDevCommandError(
                'External process failed: {0}'.format(args),
                stdout,
            )

        return stdout

    except (CalledProcessError, OSError) as e:
        raise KubeDevCommandError(
            'External process failed: {0}'.format(args),
            getattr(e, 'output', e),
        )



import re
from os import path

from setuptools import find_packages, setup


# Regex matching pattern followed by 3 numerical values separated by '.'
pattern = re.compile(r'# v(?P<version>[0-9]+\.[0-9]+(\.[0-9]+(\.[a-z0-9]+)?)?)')


def get_version():
    with open('CHANGELOG.md', 'r') as fn:
        for line in fn.readlines():
            match = pattern.fullmatch(line.strip())
            if match:
                return ''.join(match.group('version'))
    raise RuntimeError('No version found in CHANGELOG.md')


base_dir = path.abspath(path.dirname(__file__))


def get_readme_content():
    readme_file = path.join(base_dir, 'README.md')
    with open(readme_file, 'r') as f:
        return f.read()


if __name__ == '__main__':
    setup(
        version=get_version(),
        name='kubetools',
        description=(
            'Kubetools is a tool and processes for developing and deploying '
            'microservices to Kubernetes.'
        ),
        author='EDITED devs',
        author_email='dev@edited.com',
        url='http://github.com/EDITD/kubetools',
        long_description=get_readme_content(),
        long_description_content_type='text/markdown',
        packages=find_packages(),
        entry_points={
            'console_scripts': (
                # kubetools client commands
                'kubetools=kubetools.cli.__main__:main',
                # ktd dev commands
                'ktd=kubetools.dev.__main__:main',
            ),
        },
        python_requires='>=3.6',
        install_requires=(
            'click>=7,<8',
            'docker>=3,<5',
            'pyyaml>=3,<6',
            'requests>=2,<2.29.0',  # https://github.com/docker/docker-py/issues/3113
            'pyretry',
            'setuptools',
            # To support CronJob api versions 'batch/v1beta1' & 'batch/v1'
            'kubernetes>=21.7.0,<25.0.0',
            'tabulate<1',
            # compose v2 has broken container naming
            'docker-compose<2',
        ),
        extras_require={
            'dev': (
                'ipdb',
                'pytest~=6.0',
                'pytest-cov~=2.10',
                'flake8',
                'flake8-import-order',
                'flake8-commas',
            ),
        },
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Intended Audience :: Information Technology',
            'License :: OSI Approved :: MIT License',
            'Operating System :: POSIX',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Programming Language :: Python :: 3.11',
            'Programming Language :: Python :: 3.12',
            'Topic :: Software Development :: Build Tools',
            'Topic :: Software Development :: Testing',
            'Topic :: System :: Software Distribution',
        ],
    )



from os import path

from kubetools.constants import (
    GIT_BRANCH_ANNOTATION_KEY,
    GIT_COMMIT_ANNOTATION_KEY,
    GIT_TAG_ANNOTATION_KEY,
)
from kubetools.deploy.util import run_shell_command
from kubetools.exceptions import KubeBuildError


def _is_git_committed(app_dir):
    git_status = run_shell_command(
        'git', 'status', '--porcelain',
        cwd=app_dir,
    ).strip().decode()

    if git_status:
        return False
    return True


def _get_git_info(app_dir):
    git_annotations = {}

    commit_hash = run_shell_command(
        'git', 'rev-parse', '--short=7', 'HEAD',
        cwd=app_dir,
    ).strip().decode()
    git_annotations[GIT_COMMIT_ANNOTATION_KEY] = commit_hash

    branch_name = run_shell_command(
        'git', 'rev-parse', '--abbrev-ref', 'HEAD',
        cwd=app_dir,
    ).strip().decode()

    if branch_name != 'HEAD':
        git_annotations[GIT_BRANCH_ANNOTATION_KEY] = branch_name

    try:
        git_tag = run_shell_command(
            'git', 'tag', '--points-at', commit_hash,
            cwd=app_dir,
        ).strip().decode()
    except KubeBuildError:
        pass
    else:
        if git_tag:
            git_annotations[GIT_TAG_ANNOTATION_KEY] = git_tag

    return commit_hash, git_annotations


def get_git_info(app_dir, ignore_git_changes=False):
    if path.exists(path.join(app_dir, '.git')):
        if not _is_git_committed(app_dir) and not ignore_git_changes:
            raise KubeBuildError(f'{app_dir} contains uncommitted changes, refusing to deploy!')
        return _get_git_info(app_dir)
    raise KubeBuildError(f'{app_dir} is not a valid git repository!')



from os import environ, path

import click

from kubetools.exceptions import KubeDevError
from kubetools.settings import get_settings, get_settings_directory

from . import dev
from .backend import (
    get_all_containers,
    get_all_containers_by_name,
    get_container_status,
)
from .process_util import run_process


def _list_scripts(kubetools_config):
    '''
    List available dev scripts.
    '''

    settings = get_settings()
    click.echo('--> Containers & scripts:')

    for name, container in get_all_containers(kubetools_config):
        if 'devScripts' not in container:
            continue

        click.echo('    {0}:'.format(click.style(name, bold=True)))

        for script in container['devScripts']:
            if script in settings.scripts:
                click.echo('    - {0}'.format(click.style(script, 'green')))
            else:
                click.echo('    - {0} (not found)'.format(
                    click.style(script, 'red'),
                ))


@dev.command()
@click.argument('container', required=False)
@click.argument('script', required=False)
@click.pass_obj
def script(kubetools_config, container=None, script=None):
    '''
    List and execute scripts.

    The script must be made available to the container in `kubetools.yml`.
    '''

    if container is None or script is None:
        return _list_scripts(kubetools_config)

    settings = get_settings()
    config = get_all_containers_by_name(kubetools_config).get(container)

    if not config:
        raise KubeDevError('Invalid container: {0}'.format(container))

    if script not in config.get('devScripts', []):
        raise KubeDevError('Script {0} is not available in container {1}'.format(
            script, container,
        ))

    script_path = path.join(get_settings_directory(), 'scripts', script)

    if script not in settings.scripts:
        raise KubeDevError('Could not locate local script (expected: {0})'.format(
            script_path,
        ))

    status = get_container_status(kubetools_config, container)
    if not status or not status['up']:
        raise KubeDevError('Container {0} is not online'.format(container))

    script_env = environ.copy()
    script_env['APP_NAME'] = kubetools_config['name']

    # Add the envvars defined in kubetools.yml
    for env in config.get('environment', []):
        key, value = env.split('=', 1)
        script_env[key] = value

    # Add PORT_N envvars mapping the container -> host ports
    for port in status['ports']:
        # Remove the /tcp, /udp bit
        local_port = port['local'].replace('/', '')
        host_port = port['host'].split(':')[-1]
        script_env['PORT_{0}'.format(local_port)] = host_port

    # Execute the script!
    run_process([script_path], env=script_env)



class KubeError(Exception):
    type = 'generic'


# Config errors
#

class KubeConfigError(KubeError):
    type = 'config'


# Client/server errors
#

class KubeClientError(KubeError):
    type = 'client'


class KubeServerError(KubeError):
    type = 'server'


class KubeCLIError(KubeError):
    type = 'cli'


# Build errors
#

class KubeBuildError(KubeError):
    type = 'build'


# Local/dev errors
#

class KubeDevError(KubeCLIError):
    type = 'dev'


class KubeDevCommandError(KubeDevError):
    type = 'dev'



from os import path
from unittest import TestCase

from kubetools.config import load_kubetools_config
from kubetools.kubernetes.config import make_job_config


def load_job_spec():
    app_dir = path.join('tests', 'configs', 'basic_app')
    kubetools_config = load_kubetools_config(app_dir)
    return kubetools_config.get('upgrades')[0]


class TestMakeJobConfig(TestCase):
    def test_job_id_is_added_to_envvars(self):
        job_config = make_job_config(load_job_spec())
        container_env = job_config['spec']['template']['spec']['containers'][0]['env']
        self.assertIn('KUBE_JOB_ID', [env['name'] for env in container_env])

    def test_job_name_defaults_to_job_id(self):
        job_config = make_job_config(load_job_spec())
        job_name = job_config['metadata']['name']
        job_id = job_config['metadata']['labels']['job-id']
        self.assertEqual(job_name, job_id)

    def test_job_name_can_be_set_by_caller(self):
        job_config = make_job_config(load_job_spec(), job_name='myawesomejob')
        job_name = job_config['metadata']['name']
        self.assertEqual('myawesomejob', job_name)

    def test_container_name_defaults_to_upgrade(self):
        job_config = make_job_config(load_job_spec())
        container_name = job_config['spec']['template']['spec']['containers'][0]['name']
        self.assertEqual('upgrade', container_name)

    def test_container_name_can_be_set_by_caller(self):
        job_config = make_job_config(load_job_spec(), container_name='mycoolcontainer')
        container_name = job_config['spec']['template']['spec']['containers'][0]['name']
        self.assertEqual('mycoolcontainer', container_name)

    def test_resources_being_passed(self):
        expected_resources = {
            "requests": {
                "memory": "1Gi",
            },
        }
        job_config = make_job_config(load_job_spec())
        resource_config = job_config['spec']['template']['spec']['containers'][0]['resources']
        self.assertEqual(expected_resources, resource_config)

    def test_ttl_is_set_if_provided_in_config(self):
        job_spec = load_job_spec()
        ttl_option = {'ttl_seconds_after_finished': 100}
        job_spec.update(ttl_option)
        job_config = make_job_config(job_spec)
        self.assertIn('ttlSecondsAfterFinished', job_config['spec'])
        self.assertEqual(
            job_config['spec']['ttlSecondsAfterFinished'],
            ttl_option['ttl_seconds_after_finished'],
        )
